#include "toico.h"
#include <string.h>
#include <assert.h>

int isgif(u8 *s)
{
	return memcmp(s, "GIF8", 4)==0 && (s[4]=='7' || s[4] == '9');
}

struct gif_reader {
	FILE *f;
	struct img *im;
	u8 gpal[256][3]; int npal;
	int transp;

	int nblk; // remaining bytes in current img block

	int initcbits, cbits, bbits;
	u32 cbuf, cmask;

	int prefix[4096+2]; /* only >= FIRST used */
	u8 suffix[4096+2]; /* only >= FIRST used */
	int length[4096]; /* length - 1 */

	u8 stack[4096];
	int sp;

	int nxcode, clear;
};

static struct gif_reader gr;

static void bad(char *m)
{
	errx(1, "%s: %s", gr.im->name, m);
}

static inline void readn(void *p, int n)
{
	freadn(p, n, gr.f);
}

static void skip(int n)
{
	while(n--) getc(gr.f);
}

static inline u32 mkcolor(u8 c[3])
{
	return c[0]<<16 | c[1]<<8 | c[2];
}

static void read_img();

void readgif(struct img *img, FILE *f, u8 sig[SIGN_SIZE])
{
	u8 buf[5];
	u32 background;
	int stop = 0;
	int v;

	gr.f = f;
	gr.im = img;

	readn(buf, sizeof buf);
	gr.im->w = get16le(sig+6);
	gr.im->h = get16le(buf+0);

	background = ALPHA;
	gr.npal = 0;
	if(buf[2] & 0x80) {
		gr.npal = 2 << (buf[2] & 7);
		readn(gr.gpal, 3*gr.npal);
#if 0
		if(buf[3] < gr.npal)
			background = mkcolor(gr.gpal[buf[3]]);
#endif
	}

	alloc_im(img);
	{
		u32 *p = img->im, *e = IMGEND(img);
		do *p++ = background; while(p<e);
	}

	gr.transp = -1;
	for(;;) {
		int b = getc(f);
		if(b == EOF) break;
		if(b == ',') {
			read_img();
			if(stop) break;
			gr.transp = -1;
		} else if(b == '!') {
			v = getc(f);
			if(v == 0xF9) {
				readn(buf, 5);
				if(buf[0] != 4)
					bad("Wrong GCE length");
				v = buf[1] & 0x1E;
				if(v && v!=4 || buf[2] | buf[3]) stop = 1;
				if(buf[1]&1) gr.transp = buf[4];
			}
			for(;;) {
				v = getc(f);
				if(!v) break;
				skip(v);
			}
		} else
			break;
	}
}

int getcode()
{
	if(gr.cbits > gr.bbits) {
		do {
			int b;
			if(gr.nblk<=0) {
				if(gr.nblk<0)
					return -1;
				b = getc(gr.f);
				gr.nblk = -1;
				if(b<=0) break;
				gr.nblk = b;
			}
			b = getc(gr.f);
			gr.nblk--;
			if(b<0) break;
			gr.cbuf |= b<<gr.bbits;
			gr.bbits += 8;
		} while(gr.bbits<=24);
		if(gr.cbits > gr.bbits)
			return -1;
	}
	{
		int c = gr.cbuf & gr.cmask;
		gr.cbuf >>= gr.cbits;
		gr.bbits -= gr.cbits;
		return c;
	}
}

#define CLEAR (gr.clear)
#define END (gr.clear+1)
#define FIRST (gr.clear+2)

static int decode_reset()
{
	int c;
	gr.cbits = gr.initcbits;
	gr.cmask = (1<<gr.cbits)-1;
	do c = getcode(); while(c == CLEAR);
	if(c >= FIRST)
		bad("Bad code after reset");
	gr.nxcode = FIRST;
	gr.prefix[gr.nxcode] = c;
	gr.length[gr.nxcode] = 1;
	return c;
}

static void decode_start()
{
	int c;
	c = getc(gr.f);
	gr.bbits = 0;
	gr.initcbits = c+1;
	gr.clear = 1<<c;
	gr.nblk = 0;

	memset(gr.length, 0, sizeof(gr.length[0])*CLEAR);
	c = decode_reset();
	gr.sp = 1;
	gr.stack[0] = c;
}

static void inccbits()
{
	if(gr.cbits==12) return;
	gr.cbits++;
	gr.cmask = gr.cmask<<1 | 1;
}

static u32 *put_pixel(u32 *p, int i)
{
	u32 c = gr.im->pal[i];
	if(i != gr.transp) *p = c;
	return p+1;
}

static int decode(u32 *outp, int size)
{
	u32 *oute = outp + size;
	int c;

	if(gr.sp) {
from_stack:
		do {
			outp = put_pixel(outp, gr.stack[--gr.sp]);
			if(outp == oute) goto end;
		} while(gr.sp);
	}

	for(;;) {
		if(gr.nxcode>gr.cmask)
			inccbits();
		c = getcode();
		if(c<0) break;
		gr.suffix[gr.nxcode] = c;
		gr.prefix[gr.nxcode+1] = c;
		gr.length[gr.nxcode+1] = gr.length[c] + 1;
		if(c >= CLEAR) {
			if(c <= END) {
				if(c == END) break;
				c = decode_reset();
				goto out;
			} else if(c > gr.nxcode) {
				bad("Bad code");
			}

			assert(gr.prefix[c] < c);
			if(outp+gr.length[c] >= oute)
				goto do_not_fit;
			outp += gr.length[c];
			{
				int px = gr.prefix[c];
				u32 *p = outp-1;
				while(px >= CLEAR) {
					put_pixel(p--, gr.suffix[px]);
					assert(gr.prefix[px]<px);
					px = gr.prefix[px];
				}
				gr.suffix[gr.nxcode] = px;
				put_pixel(p, px);
			}
			c = gr.suffix[c];
		}
		if(gr.nxcode<4096) gr.nxcode++;
out:
		outp = put_pixel(outp, c);
		if(outp >= oute)
			break;
	}
end:
	return oute-outp;

do_not_fit:
	{
		int px = gr.prefix[c];
		int i;
		assert(gr.sp==0);
		i = 1;
		if(px >= CLEAR) do {
			gr.stack[i++] = gr.suffix[px];
			assert(gr.prefix[px]<px);
			px = gr.prefix[px];
		} while(px >= CLEAR);
		gr.stack[i] =  px;
		gr.sp = i+1;
		gr.suffix[gr.nxcode] = px;
		gr.stack[0] = gr.suffix[c];
		if(gr.nxcode<4096) gr.nxcode++;
		goto from_stack;
	}
}

static void read_img()
{
	int ix, iy, iw, ih;
	u8 ipal[256][3];
	u8 (*palette)[3];
	u8 buf[9];
	int n, i;
	
	readn(buf, 9);
	ix = get16le(buf+0);
	iy = get16le(buf+2);
	iw = get16le(buf+4);
	ih = get16le(buf+6);
	if(ix+iw > gr.im->w || iy+ih > gr.im->h)
		bad("Subimage do not fit");
	if(!(iw|ih))
		bad("No pixels in subimage");

	n = gr.npal;
	palette = gr.gpal;
	if(buf[8]&0x80) {
		n = 1<<(buf[8]&7)+1;
		readn(ipal, 3*n);
		palette = ipal;
	}
	for(i=0; i<n; i++)
		gr.im->pal[i] = mkcolor(palette[i]);

	decode_start();
	{
		int pass = 0;
		int step = gr.im->w;
		u32 *start = IMGAT(gr.im, ix, iy);
		u32 *end = IMGAT(gr.im, ix, iy+ih);
		u32 *p;
		if(buf[8]&0x40) step *= 8;
		for(p=start;;) {
			i = decode(p, iw);
			if(i) break;
			p += step;
			if(p < end) continue;

			if(step <= 2*gr.im->w) break;
			step = 4*gr.im->w >> pass++;
			p = start + step;
			step *= 2;
		}
	}

	if(gr.nblk<0) return;
	/* skip rest */
	if(gr.nblk) skip(gr.nblk);
	for(;;) { 
		n = getc(gr.f);
		if(n<=0) break;
		skip(n);
	}
}
